/* -*- linux-c -*-
 *
 * jdmouse.c: jogdial -> wheel mouse event converter
 *
 * 2002/11/13 by JR0BAK = Kunihiko IMAI <bak@d2.dion.ne.jp>
 *
 * You can freely use, modify or re-distribute this program.
 * But don't delete, modify these credit sentence.
 *
 * If you modified this program and want to re-distribute,
 * add your change log below:
 *
 * 2002/11/17: Kunihiko IMAI <bak@d2.dion.ne.jp>
 *	add built-in sound mixer control
 *	add value control with jogdial
 *
 * 2002/11/13: Kunihiko IMAI <bak@d2.dion.ne.jp>
 *      change copyright notice to easily re-distribute.
 *
 * 2001/08/23: Kunihiko IMAI <bak@d2.dion.ne.jp>
 *      previous version (I forget old change log...)
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include <string.h>
#include <sys/ioctl.h>
#include <linux/sonypi.h>

#include <linux/soundcard.h>

/* ----- configure section begin ----- */

#define SONYPI	"/dev/sonypi"		/* device file name for sonypi.o */
#define MOUSE	"/dev/jdmouse"		/* mouse FIFO file name */
#define MIXER	"/dev/mixer"		/* soundcard mixer */

#define ILLUGAL_PARTIAL_RELEASE		/* for PCG-C1XE, has invalid seq. */
#define BACKLIGHT_STEP		25
#define SOUNDVOLUME_STEP	8

#define JOGDIAL_CLICK_BUTTON1		/* left button */
//#define JOGDIAL_CLICK_BUTTON2		/* middle button */
//#define JOGDIAL_CLICK_BUTTON3		/* right button */

/*
 * command list
 */
const char *capture_half_command[] = {
	"killall -STOP mpg123",
	"killall -CONT mpg123",
	"",
};

const char *capture_deep_command[] = {
	"xset dpms force off",
	"",
};

/* ----- configure section end ----- */

//#define CYCLIC_EXAMPLE
#ifdef CYCLIC_EXAMPLE
const char *fn_1_command[] = {
	"xsetroot -solid gray",
	"xsetroot -solid seagreen4",
	"xsetroot -solid black",
	"",
};
#endif /*CYCLIC_EXAMPLE*/


int debug = 0;
int fd_mouse, fd_sonypi, fd_mixer;

#if defined(JOGDIAL_CLICK_BUTTON1)
#define MSEQ0	"\x09"
#elif defined(JOGDIAL_CLICK_BUTTON2)
#define MSEQ0	"\x0c"
#elif defined(JOGDIAL_CLICK_BUTTON3)
#define MSEQ0	"\x0a"
#else
#error define macro JOGDIAL_CLICK_BUTTON[1|2|3]
#endif


int (*jogdial_hook)( int delta ) = NULL;
const char *jogtable[] = {
	"",
	"\x08\x00\x00\x01",
	"\x08\x00\x00\xff",
	MSEQ0 "\x00\x00\x01",
	MSEQ0 "\x00\x00\xff",
	MSEQ0 "\x00\x00\x00",
	"\x08\x00\x00\x00",
};

#define DEBUG(fmt,args...) if ( debug ) fprintf ( stderr, fmt, ##args )


/*
 * cyclic_command: execute command cyclicly
 */
inline int cyclic_command ( int index, const char *cmdline[] )
{
	system ( cmdline[index] );

	if ( cmdline[++index][0] == '\0' ) {
		index = 0;
	}
	return index;
}

/*
 * soundvolume
 */
int soundvolume_prev = 0x3232;
void soundvolume_toggle ( void )
{
	int val;

	if ( fd_mixer < 0 ) return;

	ioctl ( fd_mixer, SOUND_MIXER_READ_VOLUME, &val );
	if ( val != 0 ) { /* mute */
		soundvolume_prev = val;
		val = 0;
	} else { /* unmute */
		val = soundvolume_prev;
	}
	ioctl ( fd_mixer, SOUND_MIXER_WRITE_VOLUME, &val );
}

int soundvolume_adjuster ( int delta )
{
	int val;

	if ( delta == 0 ) {
		jogdial_hook = NULL;
		return 0;
	}
	
	ioctl ( fd_mixer, SOUND_MIXER_READ_VOLUME, &val );
	DEBUG ( __FUNCTION__ ": %04x\n", val );
	val /= 256;
	val += SOUNDVOLUME_STEP * delta;
	val *= 0x101;
	if ( val < 0 ) val = 0;
	ioctl ( fd_mixer, SOUND_MIXER_WRITE_VOLUME, &val );

	return val;
}


/*
 * backlight: backlight control
 */

int backlight_adjuster ( int delta )
{
	unsigned char val, v2;
	int v;

	if ( delta == 0 ) {
		jogdial_hook = NULL;
		return 0;
	}
	do {
		v2 = val;
		ioctl ( fd_sonypi, SONYPI_IOCGBRT, &val );
	} while ( val != v2 );

	v = val + BACKLIGHT_STEP * delta;
	DEBUG ( __FUNCTION__ ":%d, %d\n", v, val );

	/* saturation */
	if ( v < 0 ) v = 0;
	if ( v > 255 ) v = 255;
	val = v;
	
	ioctl ( fd_sonypi, SONYPI_IOCSBRT, &val );
	return val;
}


void backlight_toggle ( void )
{
	unsigned char val, v2;

	do {
		v2 = val;
		ioctl ( fd_sonypi, SONYPI_IOCGBRT, &val );
	} while ( val != v2 );
	val = ( val > 128 ) ? 0 : 255;  /* toggle */
	DEBUG ( __FUNCTION__ "%d\n", val );
	ioctl ( fd_sonypi, SONYPI_IOCSBRT, &val );
}


/*
 * capture: handle capture button event
 *
 * CAPTURE button behavior:
 * 
 * o deep press and release
 *
 *	PARTIALPRESSED -> PRESSED -> PARTIALRELEASED -> RELEASED
 *
 * o half press and release
 *
 *	PARTIALPRESSED -> RELEASED
 *
 */
void capture ( int event )
{
	static int prev_event = -1;
	
	switch ( event ) {
	case SONYPI_EVENT_CAPTURE_RELEASED:
		switch ( prev_event ) {
#ifdef ILLUGAL_PARTIAL_RELEASE
		case SONYPI_EVENT_PKEY_P1:
#endif
		case SONYPI_EVENT_CAPTURE_PARTIALRELEASED: /* deep press */ {
			static int index = 0;
			index = cyclic_command ( index,
						 capture_deep_command );
			break;
		}
		case SONYPI_EVENT_CAPTURE_PARTIALPRESSED: /* half press */ {
			static int index = 0;
			index = cyclic_command ( index,
						 capture_half_command );
			break;
		}
		default:
			break;
		}
		break;
	default:
		break;
	}
	prev_event = event;
}



/*
 * jogdial: jogdial event handler
 */
void jogdial ( int event )
{
	if ( jogdial_hook != NULL ) { /* value adjuster */
		switch ( event ) {
		case SONYPI_EVENT_JOGDIAL_DOWN:
		case SONYPI_EVENT_JOGDIAL_DOWN_PRESSED:
			jogdial_hook ( -1 );
			break;
		case SONYPI_EVENT_JOGDIAL_UP:
		case SONYPI_EVENT_JOGDIAL_UP_PRESSED:
			jogdial_hook ( 1 );
			break;
		case SONYPI_EVENT_JOGDIAL_RELEASED:
			jogdial_hook ( 0 );
			break;
		default:
			break;
		}
	} else { /* mouse wheel */
		write ( fd_mouse, jogtable[event], 4 );
	}
}


void jogdial_sethook ( int (*hook_function)(int) )
{
	if ( jogdial_hook ==  hook_function ) {
		jogdial_hook = NULL;
	} else {
		jogdial_hook = hook_function;
	}
}



main ( int argc, char *argv[] )
{
	int n, o, isdeep = 0;
	char c, *sonypi = SONYPI, *mouse = MOUSE, *mixer = MIXER;
	
	while ( (o = getopt ( argc, argv, "ds:m:x:h" )) > 0 ) {
		switch ( o ) {
		case 'd':
			debug = 1;
			break;
		case 's':
			sonypi = optarg;
			break;
		case 'm':
			mouse = optarg;
			break;
		case 'x':
			mixer = optarg;
			break;
		default:
			fprintf
				( stderr,
				  "jdmouse: jogdial -> wheel mouse protocol converter\n"
				  "         2001/08/23 by JR0BAK = Kunihiko IMAI <bak@d2.dion.ne.jp>\n"
				  "\n"
				  "usage: jdmouse [-d] [-s sonypi_device] [-m mouse_FIFO] \\\n"
				  "               [-x mixer_device]\n"
				  "       -d: debug: show event ID to stdout\n"
					);
			exit ( 0 );
			break;
		}
	}
	
	
	fd_sonypi = open ( sonypi, O_RDWR );
	if ( fd_sonypi < 0 ) {
		perror ( "" );
		exit ( errno );
	}
	fd_mouse = open ( mouse, O_RDWR );
	if ( fd_mouse < 0 ) {
		perror ( "" );
		exit ( errno );
	}
	fd_mixer = open ( mixer, O_RDWR );
	if ( fd_mixer < 0 ) {
		fprintf ( stderr, "warning: cannot open %s.\n", mixer );
	}
	
	for (;;) {
		n = read ( fd_sonypi, &c, 1 );
		if ( n < 0 ) {
			break;
		}
		DEBUG ( "event: %d\n", c );
		switch ( c ) {
		case SONYPI_EVENT_JOGDIAL_DOWN:
		case SONYPI_EVENT_JOGDIAL_UP:
		case SONYPI_EVENT_JOGDIAL_DOWN_PRESSED:
		case SONYPI_EVENT_JOGDIAL_UP_PRESSED:
		case SONYPI_EVENT_JOGDIAL_PRESSED:
		case SONYPI_EVENT_JOGDIAL_RELEASED:
			jogdial ( c );
			break;
		case SONYPI_EVENT_CAPTURE_PARTIALPRESSED:
		case SONYPI_EVENT_CAPTURE_PRESSED:
		case SONYPI_EVENT_CAPTURE_RELEASED:
#ifdef ILLUGAL_PARTIAL_RELEASE
		case SONYPI_EVENT_PKEY_P1:
#endif
		case SONYPI_EVENT_CAPTURE_PARTIALRELEASED:
			capture ( c );
			break;
			
		case SONYPI_EVENT_FNKEY_F3:
			soundvolume_toggle();
			break;
		case SONYPI_EVENT_FNKEY_F4:
			jogdial_sethook( soundvolume_adjuster );
			break;
		case SONYPI_EVENT_FNKEY_F5:
			jogdial_sethook( backlight_adjuster );
			break;
		case SONYPI_EVENT_FNKEY_F6:
			backlight_toggle();
			break;
#ifdef CYCLIC_EXAMPLE
		case SONYPI_EVENT_FNKEY_1: {
			static int index = 0;

			index = cyclic_command ( index, fn_1_command );
			break;
		}
#endif
		default:
			break;
		}
	}
	exit ( 0 );
}
